import py, os, sys
import numpy as np

sys.path = [os.path.join(os.pardir, 'python')] + sys.path


class TestSNOBFIT:
    def setup_class(cls):
        import SQSnobFit, logging, numpy
        SQSnobFit.log.setLevel(logging.DEBUG)

    def reset(self):
        import SQSnobFit

     # reset the random state for each method to get predictable results
        SQSnobFit._gen_utils._randstate = np.random.RandomState(6)

    def setup_method(self, method):
        self.reset()

    def test01_simple_example(self):
        """Read access to instance public data and verify values"""

        import SQSnobFit

        def f_easy(x):
            from math import sin

            fv = np.inner(x, x)
            fv *= 1 + 0.1*sin(10*(x[0]+x[1]))

            return fv

        def run_easy(self, initial, expected):
            self.reset()

            bounds = np.array([[-1, 1], [-1, 1]], dtype=float)
            budget = 40
            x0 = np.array(initial)

            from SQSnobFit import optset
            options = optset(maxmp=2+6)

            result, history = SQSnobFit.minimize(f_easy, x0, bounds, budget, options)

          # problem is symmetric, so values may have switched: just check the sum
            assert np.round(sum(result.optpar)-sum(expected), 8) == 0
            #assert len(histout) == 10

        run_easy(self, [],         (-0.00112, -0.00078))
        run_easy(self, [0.5, 0.5], ( 0.00134, -0.00042))

    def test02_bra(self):
        """Minimize Branin's function"""

        import SQSnobFit

        def bra(x):
            from math import cos, pi

            a = 1
            b = 5.1/(4*pi*pi)
            c = 5/pi
            d = 6
            h = 10
            ff = 1/(8*pi)

            return a*(x[1]-b*x[0]**2+c*x[0]-d)**2+h*(1-ff)*cos(x[0])+h

        bounds = np.array([[-5, 5], [-5, 5]], dtype=float)
        budget = 80      # larger budget needed for full convergence
        x0 = np.array([0.5, 0.5])

        from SQSnobFit import optset
        options = optset(maxmp=len(x0)+6)

        result, history = SQSnobFit.minimize(bra, x0, bounds, budget, options)
        # LIMIT:
        # fglob = 0.397887357729739
        # xglob = [3.14159265, 2.27500000]
        assert np.round(sum(result.optpar)-sum((3.1416, 2.275)), 8) == 0

    def test03_Hartman6(self):
        """Minimize Hartman6 function"""

        import SQSnobFit

        def Hartman6(x):
            import numpy, math

            a = numpy.array(
                [[10.00,  0.05,  3.00, 17.00],
                 [ 3.00, 10.00,  3.50,  8.00],
                 [17.00, 17.00,  1.70,  0.05],
                 [ 3.50,  0.10, 10.00, 10.00],
                 [ 1.70,  8.00, 17.00,  0.10],
                 [ 8.00, 14.00,  8.00, 14.00]])

            p = numpy.array(
                [[0.1312, 0.2329, 0.2348, 0.4047],
                 [0.1696, 0.4135, 0.1451, 0.8828],
                 [0.5569, 0.8307, 0.3522, 0.8732],
                 [0.0124, 0.3736, 0.2883, 0.5743],
                 [0.8283, 0.1004, 0.3047, 0.1091],
                 [0.5886, 0.9991, 0.6650, 0.0381]])

            c = numpy.array([1.0, 1.2, 3.0, 3.2])

            d = numpy.zeros((4,))
            for i in range(4):
                d[i] = sum(a[:,i]*(x - p[:,i])**2)

            return -(c.dot(numpy.exp(-d)))

        def run_Hartman6(self, initial, expected):
            self.reset()

            bounds = np.array([[0, 1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1]], dtype=float)
            budget = 250
            x0 = np.array(initial)

            from SQSnobFit import optset
            options = optset(maxmp=6+6)
            result, history = SQSnobFit.minimize(Hartman6, x0, bounds, budget, options)

            assert np.round(sum(result.optpar)-sum(expected), 8) == 0

      # note: results are still subtly different from the reference, but the most likely
      # reason seems to be that the respective random number generators differ slightly
        run_Hartman6(self, [],      (0.2077,  0.14892, 0.4829,  0.2725,  0.31493, 0.66138))
        run_Hartman6(self, [0.5]*6, (0.20133, 0.1504,  0.47666, 0.27787, 0.3134,  0.65797))

    def test04_direct_call(self):
        """Direct call of a single iteration"""

        import SQSnobFit
        import numpy as np

        x = np.array([[23, 23], [ 50, 50],  [50, 70],  [70, 70]])
        f = np.array([[ 0, -1], [-34, -1], [-83, -1], [-85, -1]])

        bounds = np.array([[0, 100], [0, 100]])
        config = {'bounds': bounds, 'p': .5, 'nreq': 2*2+6}
        dx = (bounds[:,1]-bounds[:,0])*1E-2

        request, xbest, fbest = SQSnobFit.snobfit(x,f,config,dx)

        assert xbest[0] == 70 and xbest[1] == 70
        assert fbest == -85



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to orse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES
# LOSS OF USE, DATA, OR PROFITS OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function xl, xu, x, f, nsplit, small = snobsplit(x, f, xl0, xu0, nspl, u, v)

 Splits a box [xl0,xu0] contained in a bigger box [u,v] such that each of the
 resulting boxes contains just one point of a given set of points.

 Input:
  x             the rows are a set of points
  f - f[j,:]    contains the function value, its variation and possibly
                other parameters belonging to point x(j,:)
  xl0           vector of lower bounds of the box
  xu0           vector of upper bounds of the box
  nspl          nspl[i] is the number of splits the box [xl0,xu0] has already
                undergone along the ith coordinate
  u, v          bounds of the original problem; [xl0, xu0] is contained in [u, v]

 Output:
  xl - xl[j,:]  is the lower bound of box j
  xu - xi[j,:]  is the upper bound of box j
  x - x[j,:]    is the point contained in box j
  f - f[j,:]    contains the function value at x(j,:), its uncertainty etc.
  nsplit - nsplit[j,i] is the number of times box j has been split in the
                ith coordinate
  small - small[j] integer-valued logarithmic volume measure of box j
"""

from ._gen_utils import rsort, max_, find, std
import numpy


def snobsplit(x, f, xl0, xu0, nspl=None, u=None, v=None):
    n = x.shape[1]
    if len(xl0) > 1:
        xl0 = xl0.T
        xu0 = xu0.T

    if nspl is None or nspl.size <= 0:
        nspl = numpy.zeros((1, n))

    if u is None or u.size <= 0:
        u = xl0.copy()
    if v is None or v.size <= 0:
        v = xu0.copy()

    if len(x) == 1:
        xl = xl0.copy()
        xu = xu0.copy()
        nsplit = nspl.copy()
        small = numpy.array([-numpy.sum(numpy.round(numpy.log2((xu-xl)/(v-u))))])
        return xl, xu, x, f, nsplit, small

    elif len(x) == 2:
        dmax, i = max_((numpy.abs(x[0]-x[1])/(v-u)).flatten())
        ymid = 0.5*(x[0,i]+x[1,i])
        xl = numpy.zeros((2, n))
        xu = numpy.zeros((2, n))
        xl[0] = xl0
        xu[0] = xu0
        xl[1] = xl0
        xu[1] = xu0
        if x[0,i] < x[1,i]:
            xu[0,i] = ymid
            xl[1,i] = ymid
        else:
            xl[0,i] = ymid
            xu[1,i] = ymid

        nsplit = numpy.zeros((2, n))
        nsplit[0] = nspl
        nsplit[0,i] = nsplit[0,i] + 1
        nsplit[1,i] = nsplit[0,i]
        small = -numpy.sum(numpy.round(numpy.log2((xu-xl)/(numpy.ones((2,1))*(v-u)))),1).T
        return xl, xu, x, f, nsplit, small

    var = numpy.zeros(v.shape[1])

    for i in range(n):
        var[i] = std(x[:,i]/(v.T[i]-u.T[i]))

    xx = numpy.sort(x,0)
    dd = xx[1:xx.shape[0]]-xx[0:xx.shape[0]-1]
    mvar, i = max_(var)
    y, w, cdf, dof = rsort(x[:,i], remove_dups=False)

    d = y[1:y.shape[0]] - y[0:y.shape[0]-1]
    ld = len(d)

    ii = range(ld)
    dmax, j = max_(d[ii])
    j = ii[j]

    ymid = 0.5*(y[j] + y[j+1])
    ind1 = find(x[:,i] < ymid)
    ind2 = find(x[:,i] > ymid)

    xl = numpy.zeros((2, n))
    xu = numpy.zeros((2, n))
    nsplit = numpy.zeros((2, n))
    npoint = numpy.zeros(2)
    ind = numpy.zeros((2, max(len(ind1), len(ind2))), dtype=int)
    xl[0] = xl0
    xu[0] = xu0
    xu[0,i] = ymid
    xl[1] = xl0
    xl[1,i] = ymid
    xu[1] = xu0
    nsplit[0] = nspl
    nsplit[0,i] = nsplit[0,i] + 1
    nsplit[1] = nsplit[1,:]
    npoint[0] = len(ind1)
    npoint[1] = len(ind2)
    # Note: adding 1 here, so that later we can search for 0 as "False"
    ind[0,:len(ind1)] = ind1.T+1
    ind[1,:len(ind2)] = ind2.T+1
    nboxes = 1

    maxpoint, j = max_(npoint)
    while maxpoint > 1: 
        ind0 = ind[j,find(ind[j,:])].astype(int)
        # Undo the add of 1 to make these values back to indices again
        ind0 -= 1
        for i_ in range(n):
            while i_ >= len(var):
                var = numpy.append(var, 0)
            var[i_] = std(x[ind0,i_]/(v.T[i_]-u.T[i_]))

        maxvar, i = max_(var)
        rsort_x = (x[ind0,i])[:,0]
        y, w, cdf, dof = rsort(rsort_x, remove_dups=False)

        d = y[1:len(y)] - y[0:len(y)-1]
        ld = len(d)

        ii = numpy.arange(ld).astype(int)
        dmax, k = max_(d[ii])
        k = ii[k]
        ymid = 0.5*(y[k]+y[k+1])
        ind1 = find(x[ind0,i]<ymid)
        ind2 = find(x[ind0,i]>ymid)
        ind1 = ind0[ind1[:,0],0]
        ind2 = ind0[ind2[:,0],0]
        nboxes = nboxes + 1

        if len(xl) <= nboxes:
            xl = numpy.append(xl, numpy.zeros((len(xl)-nboxes+1, xl.shape[1])), axis=0)
        if len(xu) <= nboxes:
            xu = numpy.append(xu, numpy.zeros((len(xu)-nboxes+1, xu.shape[1])), axis=0)
        if len(nsplit) <= nboxes:
            nsplit = numpy.append(nsplit, numpy.zeros((len(nsplit)-nboxes+1, nsplit.shape[1])), axis=0)

        while nboxes >= len(npoint):
            npoint = numpy.append(npoint, 0)
        while nboxes >= ind.shape[0]:
            ind = numpy.append(ind, [numpy.zeros(ind.shape[1])], axis=0)

        xl[nboxes] = xl[j]
        xu[nboxes] = xu[j]
        xu[j,i] = ymid
        xl[nboxes,i] = ymid
        nsplit[j,i] = nsplit[j,i] + 1
        nsplit[nboxes,:] = nsplit[j,:]
        npoint[j] = len(ind1)
        npoint[nboxes] = len(ind2)
        ind[j,:ind.shape[1]] = 0
        # Note: here, too, adding 1, so that later we can search for 0 as "False"
        ind[j,:len(ind1)] = ind1+1
        ind[nboxes,:len(ind2)] = ind2+1
        if not numpy.sum(ind[:,ind.shape[1] - 1]):
            ind = ind[:,:ind.shape[1] - 1]
        maxpoint, j = max_(npoint)

    if ind.size <= 0:
        x = numpy.array([None])
        f = numpy.array([None])
    else:
        ind = ind.astype(int)
        x = x[ind[:,0]-1]
        f = f[ind[:,0]-1]
    small = -numpy.sum(numpy.round(numpy.log2((xu-xl)/(numpy.ones((x.shape[0],1))*(v-u)))),1).T

    return xl, xu, x, f, nsplit, small



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to orse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function y, f, g, sigma = snoblocf(j, x, f, near, dx, u, v)

 Computes a local fit around the point x0 = x(j,:) and minimizes it
 on a trust region.
 
 Input:
  j             index of the point around which the fit is to be computed
  x             the rows contain the points where the function has been
                evaluated
  f             the corresponding function values and their uncertainties,
                i.e., f[j,0] = f[x[j,:]] and f[j,1] = df[x[j,:]]
  near          near[j,:] is a vector containing the indices of the nearest
                neighbors of the point x[j,:]
  dx            resolution vector, i.e. the ith coordinate of a point to be
                generated is an integer-valued multiple of dx(i)
  u, v          bounds of the box where the points should be generated
 
 Output:
  y             estimated minimizer in the trust region
 
  f1            its estimated function value
  g             estimated gradient for the fit
  sigma         sigma = norm(A*g-b)/sqrt(K-n), where A and b are the
                coefficients resp. right hand side of the fit, n is the
                dimension and K the number of nearest neighbors considered
                (estimated standard deviation of the model errors)
"""

from ._gen_utils import diag, rsort, max_, find, std, maximum_, rand
import numpy


def snoblocf(j, x, f, near, dx, u, v):
    n = u.shape[1]  # dimension of the problem
    x0 = x[j,:]
    f0 = f[j,0]

    df0 = f[j,1]
    D = df0/dx**2
    x1 = x[near[j].astype(int)]
    K = len(x1)
    S = x1-numpy.outer(numpy.ones(K), x0)
    d = 0.5*numpy.abs(S).max(0)
    d = numpy.maximum(d,dx)
    sc = numpy.zeros((1, K))
    for i in range(K):
        sc[0,i] = S[i].dot(diag(D).dot(S[i].T)) + f[int(near[j,i]),1]

    A = S / (sc.T.dot(numpy.ones((1,n))) )
    b = (f[near[j].astype(int),0].reshape(near.shape[1], 1) - f0) / (sc.T)
    U, Sigma, V = numpy.linalg.svd(A, 0)
    V = V.T
    Sigma = diag(Sigma)
    Sigma = numpy.amax(Sigma, axis=1)

    g = V.dot(diag(1./Sigma).dot((U.T).dot(b))).flatten()
    sigma = numpy.sqrt(numpy.sum((A.dot(g)-b.T)**2)/(K-n))

    pl = numpy.maximum(-d, u.flatten()-x0)
    pu = numpy.minimum(d, v.flatten()-x0)
    p = numpy.zeros((n,))
    for i in range(n):
        p[i] = snobqmin(sigma*D[i], g[i], pl[i], pu[i])

    y = snobround(x0+p, u, v, dx)
    nc = 0

    while (numpy.min(numpy.maximum(numpy.abs(x - numpy.outer(numpy.ones(len(x)), y) -
            numpy.outer(numpy.ones(len(x)), dx)), 1)) < 0) and nc < 5:
        p = pl + (pu-pl)*rand(1, n)
        y = snobround(x0+p, u, v, dx)
        nc = nc + 1

    p = y - x0
    err = p.dot(diag(D).dot(p.T)) + df0
    f1 = f0 + p.dot(g) + sigma*err

    return y, f1, g, sigma


#-----
def snobqmin(a, b, xl, xu):
    """
      function x = snobqmin(a, b, xl, xu)

      Minimization of the quadratic polynomial p(x) = a*x^2+b*x over [xl, xu].

      Input:
      a, b      coefficients of the polynomial
      xl, xu    bounds (xl < xu)

      Output:
      x         minimizer in [xl, xu]
    """

    if a > 0:
        x = -0.5*b/a
        x = numpy.minimum(numpy.maximum(xl, x), xu)
    else:
        fl = a*xl**2+b*xl
        fu = a*xu**2+b*xu
        if (fu <= fl).any():
            x = xu
        else:
            x = xl

    return x


#-----
def snobround(x, u, v, dx):
    """
      function x = snobround(x, u, v, dx)

      A point x is projected into the interior of [u, v] and x[i] is
      rounded to the nearest integer multiple of dx[i].

      Input:
      x         vector of length n
      u, v      vectors of length n such that u < v
      dx        vector of length n

      Output:
      x         projected and rounded version of x
    """

    x = numpy.minimum(numpy.maximum(x, u), v)
    dx = dx.reshape(x.shape)
    x = numpy.multiply(numpy.round(numpy.divide(x, dx)), dx)

    i1 = find(x<u)
    for i in i1:
        x[0, i1] += dx[0, i1]

    i2 = find(x>v)
    for i in i2:
        x[0, i2] -= dx[0, i2]

    return x



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to orse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function x5 = snob5(x, u, v, dx, nreq)

 Generates nreq points of class 5 in [u, v].
 
 Input:
  x             the rows are the points already chosen by Snobfit (can be
                empty)
  u, v          bounds of the box in which the points are to be generated
  dx            resolution vector, i.e. the ith coordinate of a point to be
                generated is an integer-valued multiple of dx(i)
  nreq          number of points to be generated
 
 Output:
  x5            the rows are the requested points
                x5 is of dimension nreq x n (n = dimension of the problem)
"""
  
from ._gen_utils import find, max_, rand
from ._snoblocf  import snobround
import numpy


def snob5(x, u, v, dx, nreq):
    n = u.shape[1] # dimension of the problem
    nx = len(x)    # number of data points
    nx5 = 100*nreq
    d = numpy.zeros(nx5)
    xnew = numpy.outer(numpy.ones(nx5), u) + rand(nx5,n)*numpy.outer(numpy.ones(nx5), v-u)
    if nx:
        for j in range(nx5):
            xnew[j,:] = snobround(xnew[j,:], u, v, dx)
            d[j] = numpy.min(numpy.sum((x-numpy.ones((nx,1))*xnew[j,:])**2, 1))

        ind = find(d == 0)
        xnew = numpy.delete(xnew, ind, 0)
        d = numpy.delete(d, ind, 0)
        x5 = numpy.array([]).reshape(0, n)
        nx5 = len(xnew)
    else:
        x5 = xnew[0,:]
        xnew = xnew[1:]
        nx5 = nx5-1
        d = numpy.sum((xnew - numpy.outer(numpy.ones(nx5), x5))**2, axis=1, keepdims=True).flatten()
        nreq = nreq - 1

    for j in range(nreq):
        if d.size <= 0:
            break
        dmax, i = max_(d)
        y = xnew[i,:]
        x5 = numpy.vstack((x5, y))
        xnew = numpy.delete(xnew, i, 0)
        d = numpy.delete(d, i, 0)
        nx5 = nx5 - 1
        d1 = numpy.sum((xnew - numpy.outer(numpy.ones(nx5), y))**2, 1)
        d = numpy.minimum(d, d1)

    return x5



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to orse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function local, nlocal = snoblp(f, near, ind)

 Computes a pointer to all "local" points (i.e. points whose neighbors have
 "significantly larger" function values).
 
 Input:
  f            f[j] is the function value of point j
  near         near[j,:] contains the indices of the near.shape[1] neighbors of
 	       point j
  ind          pointer to the boxes to be considered (optional, default 0:len(f)))
 
 Output:
  local        vector containing the indices of all local points
  nlocal       vector containing the indices of all nonlocal points
"""

import numpy


def snoblp(f, near, ind=None):
    if ind is None:
        ind = numpy.arange(len(f))

    local = numpy.array([], dtype=int)
    nlocal = ind.T
    jj = numpy.array([], dtype=int)
    for j in range(len(ind)):
        fmi = numpy.min(f[near[ind[j]]])
        fma = numpy.max(f[near[ind[j]]])
        if f[ind[j]] < fmi - 0.2*(fma-fmi):
            local = numpy.append(local, ind[j])
            jj = numpy.append(jj, j)

    if len(jj):
        nlocal = numpy.delete(nlocal, jj)

    nlocal = nlocal.flatten()
    return local, nlocal



from __future__ import print_function
# Python version of SNOBFIT v2.1 "snobfit.m" MATLAB version by A. Neumaier.
#
# Modified and redistributed with permission.

# Original copyright and license notice:
#
# Copyright (c) 2003-2008, Arnold Neumaier
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of Vienna nor the
#       names of its contributors may be used to orse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ARNOLD NEUMAIER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL  ARNOLD NEUMAIER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
 function f = snobnan(fnan, f, near, inew)

 Replaces the function values NaN of a set of points by a value determined by
 their nearest neighbors with finite function values, with a safeguard for the
 case that all neighbors have function value NaN.
 
 Input:
  fnan         vector containing the pointers to the points where the
               function value could not be obtained
  f            f[:,0] set of available function values
               f[:,1] their uncertainty/variation
  near[j,:]    vector pointing to the nearest neighbors of point j
  inew         vector pointing to the new boxes and boxes whose nearest
               neighbors have changed
 
 Output:
  f            updated version of f
"""

from ._gen_utils import rsort, max_, min_, find, std
import numpy


def snobnan(fnan, f, near, inew):
    lnn = near.shape[1]
    fnan = fnan.flatten()
    notnan = filter(lambda x: x not in fnan, range(len(f)))
    fmx, imax = max_(f[notnan,0])
    fmn, imin = min_(f[notnan,0])
    dfmax = f[imax,1]

    for l in fnan:
        if find(inew==l).size > 0:
            # a substitute function value is only computed for new points and for
            # points whose function values have changed
            ind = near[l,:]

            # eliminate neighbors with function value NaN
            ind1 = numpy.array([])
            for i in range(len(ind)):
                 if (find(fnan == ind[i])).size > 0:
                     ind1 = numpy.concatenate((ind1, i), 0)

            if ind1.size > 0:
                ind = numpy.delete(ind, ind1, 0)

            if ind.size <= 0:
                f[l,0] = fmx + 1.e-3*(fmx-fmn)
                f[l,1] = dfmax
            else:
                 fmax1, k = max_(f[ind, 0])
                 fmin1, temp = min_(f[ind, 1])
                 f[l,0] = fmax1 + 1.e-3*(fmax1-fmin1)
                 f[l,1] = f[ind[k],1]

    return f
